package russotto.jglulx;
import glk.GlkDispatcher;
import glk.GlkArgument;
public class GInstruction extends GInstrGen
{
    public final static int OPMODE_0 = 0;
    public final static int OPMODE_BYTE = 1;
    public final static int OPMODE_SHORT = 2;
    public final static int OPMODE_WORD = 3;

    public final static int OPMODE_BYTE_ADDR = 5;
    public final static int OPMODE_SHORT_ADDR = 6;
    public final static int OPMODE_WORD_ADDR = 7;
    public final static int OPMODE_STACK = 8;
    public final static int OPMODE_BYTE_LOCAL = 9;
    public final static int OPMODE_SHORT_LOCAL = 0xA;
    public final static int OPMODE_WORD_LOCAL = 0xB;

    public final static int OPMODE_BYTE_RAM = 0xD;
    public final static int OPMODE_SHORT_RAM = 0xE;
    public final static int OPMODE_WORD_RAM = 0xF;

    public final static int DESTTYPE_DISCARD = 0;
    public final static int DESTTYPE_MEMORY = 1;
    public final static int DESTTYPE_LOCAL = 2;
    public final static int DESTTYPE_STACK = 3;

    protected int nstores;
    protected int nloads;
    protected int store_cursor;
    protected int noperands;
    protected int op_width;
    protected int opindirect[] = new int[MAXOPERANDS];
    Glulx gm;

    GInstruction(Glulx gm)
    {
	this.gm = gm;
    }

    public String load_operand_string(int i)
    {
	String opstring = "(oops)" + Integer.toHexString(opmodes[i]);
	switch (opmodes[i]) {
	case OPMODE_0:
	    opstring = "0";
	    break;
	case OPMODE_BYTE:
	case OPMODE_WORD:
	case OPMODE_SHORT:
	    opstring = Integer.toHexString(opvalues[i]);
	    break;
	case OPMODE_BYTE_RAM:
	case OPMODE_WORD_RAM:
	case OPMODE_SHORT_RAM:
	    opstring = "R";
	    opstring += "(" + Integer.toHexString(opindirect[i]) + ")";
	    opstring += "(=" + Integer.toHexString(opvalues[i])+")";
	case OPMODE_BYTE_ADDR:
	case OPMODE_WORD_ADDR:
	case OPMODE_SHORT_ADDR:
	    opstring = "(" + Integer.toHexString(opindirect[i]) + ")";
	    opstring += "(=" + Integer.toHexString(opvalues[i])+")";
	    break;
	case OPMODE_BYTE_LOCAL:
	case OPMODE_WORD_LOCAL:
	case OPMODE_SHORT_LOCAL:
	    opstring = "l@" + Integer.toHexString(opindirect[i]);
	    opstring += "(=" + Integer.toHexString(opvalues[i])+")";
	    break;
	case OPMODE_STACK:
	    opstring = "(-sp)";
	    opstring += "(=" + Integer.toHexString(opvalues[i])+")";
	    break;
	}
	return opstring;
    }

    public String store_operand_string(int i)
    {
	String opstring = "<oops>" + Integer.toHexString(opmodes[i]);
	switch (opmodes[i]) {
	case OPMODE_0:
	    opstring = "<discard>";
	    break;
	case OPMODE_BYTE_ADDR:
	case OPMODE_WORD_ADDR:
	case OPMODE_SHORT_ADDR:
	    opstring = "<" + Integer.toHexString(opvalues[i]) + ">";
	    break;
	case OPMODE_BYTE_RAM:
	case OPMODE_WORD_RAM:
	case OPMODE_SHORT_RAM:
	    opstring = "R<" + Integer.toHexString(opvalues[i]) + ">";
	    break;
	case OPMODE_BYTE_LOCAL:
	case OPMODE_WORD_LOCAL:
	case OPMODE_SHORT_LOCAL:
	    opstring = "<l@" + Integer.toHexString(opvalues[i]) + ">";
	    break;

	case OPMODE_STACK:
	    opstring = "<sp+>";
	}
	return opstring;
    }

    public void log_instruction(int pc)
    {
	int i;
	String logme;

	if (!Log.willLog(Log.DEBUG_EXTREME, "log_instruction")) return;
	logme = Integer.toHexString(pc) + ": ";
	switch (logme.length()) {
	case 8: logme = " " + logme; break;
	case 7: logme = "  " + logme; break;
	case 6: logme = "   " + logme; break;
	case 5: logme = "    " + logme; break;
	case 4: logme = "     " + logme; break;
	case 3: logme = "      " + logme; break;
	case 2: logme = "       " + logme; break;
	case 1: logme = "        " + logme; break;
	}
	logme += opcodes[opcode].name;
	for (i = 0; i < nloads; i++) {
	    logme += "\t";
	    logme += load_operand_string(i);
	}
	for (i = nloads; i < noperands; i++) {
	    logme += "\t";
	    logme += store_operand_string(i);
	}
	Log.log(Log.DEBUG_EXTREME, "log_instruction", logme);
    }

    protected void decode_opcode() 
    {
	opcode = gm.getCodeByte();
	if ((opcode & 0xC0) == 0xC0) {
	    opcode = ((opcode&0x3F)<<24) | (gm.getCodeByte() << 16);
	    opcode |= gm.getCodeShort();
	}
	else if ((opcode & 0xC0) == 0x80) {
	    opcode = ((opcode&0x3F)<<8) | gm.getCodeByte();
	}
    }

    int second_nybble;
    boolean inside_byte = false;
    int getCodeNybble()
    {
	int the_byte;
	if (!inside_byte) {
	    the_byte = gm.getCodeByte();
	    inside_byte = true;
	    second_nybble = the_byte >>> 4;
	    return (the_byte & 0xF);
	}
	inside_byte = false;
	return second_nybble;
    }

    void reset_nybble()
    {
	inside_byte = false;
    }
    
    public void decode_instruction() throws GlulxException
    {
	decode_opcode();
	GOpcode op = opcodes[opcode];
	op_width = op.op_width;
	noperands = (nloads = op.load_operands);
	noperands += (nstores = op.store_operands);
	store_cursor = nloads;
	for (int i = 0; i < noperands; i++) {
	    opmodes[i] = getCodeNybble();
	}
	reset_nybble();
	decode_loads();
	decode_stores();
	// Assumes stores strictly follow loads
    }

    public void decode_loads() throws GlulxException 
    {
	long indirect_addr = 0;
	
	for (int i = 0; i < nloads; i++) {
	    switch(opmodes[i]) {
	    case OPMODE_0:
		opvalues[i] = 0;
		break;
	    case OPMODE_BYTE:
		opvalues[i] = (byte)gm.getCodeByte(); // the cast is necessary
		break;
	    case OPMODE_SHORT:
		opvalues[i] = (short)gm.getCodeShort(); // the cast is necessary
		break;
	    case OPMODE_WORD:
		opvalues[i] = gm.getCodeWord();
		break;
		
	    case OPMODE_BYTE_ADDR:
	    case OPMODE_BYTE_RAM:
	    case OPMODE_BYTE_LOCAL:
		indirect_addr = gm.getCodeByte();
		break;
		
	    case OPMODE_SHORT_RAM:
	    case OPMODE_SHORT_ADDR:
	    case OPMODE_SHORT_LOCAL:
		indirect_addr = gm.getCodeShort();
		break;
		
	    case OPMODE_WORD_RAM:
	    case OPMODE_WORD_LOCAL:
	    case OPMODE_WORD_ADDR:
		indirect_addr = gm.getCodeWord()&0xFFFFFFFF;
		break;
		
	    case OPMODE_STACK:
		opvalues[i] = gm.g_stack.popWord();
		break;
		
	    default:
		throw new GlulxException("Bogus Addressing Mode " + opmodes[i]);
	    }

	    opindirect[i] = (int)indirect_addr;

	    switch(opmodes[i]) {
	    case OPMODE_BYTE_LOCAL:
	    case OPMODE_SHORT_LOCAL:
	    case OPMODE_WORD_LOCAL:
		switch (op_width) {
		case 4:
		    opvalues[i] = gm.readLocalWord(indirect_addr);
		    break;
		case 2:
		    opvalues[i] = gm.readLocalShort(indirect_addr);
		    break;
		case 1:
		    opvalues[i] = gm.readLocalByte(indirect_addr);
		    break;
		}
		break;

	    case OPMODE_BYTE_RAM:
	    case OPMODE_SHORT_RAM:
	    case OPMODE_WORD_RAM:
		indirect_addr = (indirect_addr + gm.RAMSTART) & 0xFFFFFFFF;
		opindirect[i] = (int)indirect_addr;
	    case OPMODE_BYTE_ADDR:
	    case OPMODE_SHORT_ADDR:
	    case OPMODE_WORD_ADDR:
		switch (op_width) {
		case 4:
		    opvalues[i] = gm.g_memory.readWord(indirect_addr);
		    break;
		case 2:
		    opvalues[i] = gm.g_memory.readShort(indirect_addr);
		    break;
		case 1:
		    opvalues[i] = gm.g_memory.readByte(indirect_addr);
		    break;
		}
		break;
	    }
	}
    }
  
    public void do_store(int val) throws GlulxException
    {
	if (store_cursor >= noperands)
	    throw new GlulxException("Tried to store too many values");
	gm.do_store(opmodes[store_cursor], opvalues[store_cursor], op_width, val);
	store_cursor++;
    }

    public void decode_stores() throws GlulxException
    {
	for (int i = nloads; i < noperands; i++) {
	    switch(opmodes[i]) {
	    case OPMODE_BYTE_ADDR:
	    case OPMODE_BYTE_RAM:
	    case OPMODE_BYTE_LOCAL:
		opvalues[i] = gm.getCodeByte();
		break;

	    case OPMODE_SHORT_ADDR:
	    case OPMODE_SHORT_RAM:
	    case OPMODE_SHORT_LOCAL:
		opvalues[i] = gm.getCodeShort();
		break;

	    case OPMODE_WORD_ADDR:
	    case OPMODE_WORD_RAM:
	    case OPMODE_WORD_LOCAL:
		opvalues[i] = gm.getCodeWord();
		break;

	    case OPMODE_0:
	    case OPMODE_STACK:
		opvalues[i] = 0;
		break;

	    default:
		throw new GlulxException("Bogus Addressing Mode for Store" + opmodes[i]);
		
	
	    }
	    switch(opmodes[i]) {
	    case OPMODE_SHORT_RAM:
	    case OPMODE_BYTE_RAM:
	    case OPMODE_WORD_RAM:
		opvalues[i] += gm.RAMSTART;
	    }
	}
    }

    public static int getModeFromDestType(int desttype) 
    {
	switch (desttype) {
	case DESTTYPE_MEMORY:
	    return OPMODE_WORD_ADDR;
	case DESTTYPE_LOCAL:
	    return OPMODE_WORD_LOCAL;
	case DESTTYPE_STACK:
	    return OPMODE_STACK;
	case DESTTYPE_DISCARD:
	    return OPMODE_0;
	}
	return -1;
    }

    public static int getDestType(int addrmode) 
    {
	    switch(addrmode) {
	    case OPMODE_BYTE_ADDR:
	    case OPMODE_WORD_RAM:
	    case OPMODE_BYTE_RAM:
	    case OPMODE_SHORT_ADDR:
	    case OPMODE_SHORT_RAM:
	    case OPMODE_WORD_ADDR:
		return DESTTYPE_MEMORY;

	    case OPMODE_BYTE_LOCAL:
	    case OPMODE_SHORT_LOCAL:
	    case OPMODE_WORD_LOCAL:
		return DESTTYPE_LOCAL;

	    case OPMODE_0:
		return DESTTYPE_DISCARD;

	    case OPMODE_STACK:
		return DESTTYPE_STACK;
	    }
	    return -1;
    }
    
    public void op_call() throws GlulxException
    {
	int nargs = opvalues[1];
	int args[] = new int[nargs];
	Log.log(Log.DEBUG, "op_call",
		"Calling: " + Integer.toHexString(opvalues[0])
		            + " fp = " + Integer.toHexString(gm.fp));
	for (int i = 0; i < nargs; i++) {
	    args[i] = gm.g_stack.popWord();
	    Log.log(Log.DEBUG, "op_call", "args[" + i + "] = " +
			       Integer.toHexString(args[i]));
	}

	gm.make_callframe(getDestType(opmodes[2]), opvalues[2]);
	gm.pc = opvalues[0];
	gm.make_callstack(args);
    }

    public void op_callf() throws GlulxException
    {
	int args[] = new int[0];

	Log.log(Log.DEBUG, "op_callf", "Calling: " + Integer.toHexString(opvalues[0])
			   + " fp = " + Integer.toHexString(gm.fp));
	gm.make_callframe(getDestType(opmodes[1]), opvalues[1]);
	gm.pc = opvalues[0];
	gm.make_callstack(args);
    }

    public void op_callfi() throws GlulxException
    {
	int args[] = new int[1];

	args[0] = opvalues[1];
        Log.log(Log.DEBUG, "op_callfi", 
		"Calling: " + Integer.toHexString(opvalues[0])
			   + " fp = " + Integer.toHexString(gm.fp));
	Log.log(Log.DEBUG, "op_callfi", "args[0] = " +  Integer.toHexString(args[0]));
	gm.make_callframe(getDestType(opmodes[2]), opvalues[2]);
	gm.pc = opvalues[0];
	gm.make_callstack(args);
    }

    public void op_callfii() throws GlulxException
    {
	int args[] = new int[2];

	args[0] = opvalues[1];
	args[1] = opvalues[2];
	Log.log(Log.DEBUG, "op_callfii", "Calling: " + Integer.toHexString(opvalues[0])
			   + " fp = " + Integer.toHexString(gm.fp));
	Log.log(Log.DEBUG, "op_callfii", "args[0] = " +  Integer.toHexString(args[0]));
	Log.log(Log.DEBUG, "op_callfii", "args[1] = " +  Integer.toHexString(args[1]));
	gm.make_callframe(getDestType(opmodes[3]), opvalues[3]);
	gm.pc = opvalues[0];
	gm.make_callstack(args);
    }

    public void op_callfiii() throws GlulxException
    {
	int args[] = new int[3];

	args[0] = opvalues[1];
	args[1] = opvalues[2];
	args[2] = opvalues[3];
	Log.log(Log.DEBUG, "op_callfiii", "Calling: " + Integer.toHexString(opvalues[0])
			   + " fp = " + Integer.toHexString(gm.fp));
	Log.log(Log.DEBUG, "op_callfiii", "args[0] = " +  Integer.toHexString(args[0]));
	Log.log(Log.DEBUG, "op_callfiii", "args[1] = " +  Integer.toHexString(args[1]));
	Log.log(Log.DEBUG, "op_callfiii", "args[2] = " +  Integer.toHexString(args[2]));
	gm.make_callframe(getDestType(opmodes[4]), opvalues[4]);
	gm.pc = opvalues[0];
	gm.make_callstack(args);
    }

    public void op_return() throws GlulxException
    {
	gm.do_return(opvalues[0]);
    }

    public void op_copy() throws GlulxException
    {
	do_store(opvalues[0]);
    }
    
    public void op_copys() throws GlulxException
    {
	do_store(opvalues[0]&0xFFFF);
    }
    
    public void op_copyb() throws GlulxException
    {
	do_store(opvalues[0]&0xFF);
    }
    
    public void op_add() throws GlulxException
    {
	do_store(opvalues[0] + opvalues[1]);
    }

    public void op_sub() throws GlulxException
    {
	do_store(opvalues[0] - opvalues[1]);
    }

    public void op_mul() throws GlulxException
    {
	do_store(opvalues[0] * opvalues[1]);
    }

    public void op_div() throws GlulxException
    {
	do_store(opvalues[0] / opvalues[1]);
    }

    public void op_mod() throws GlulxException
    {
	do_store(opvalues[0] % opvalues[1]);
    }

    public void op_neg() throws GlulxException
    {
	do_store(-opvalues[0]);
    }

    public void op_bitand() throws GlulxException
    {
	do_store(opvalues[0] & opvalues[1]);
    }

    public void op_bitor() throws GlulxException
    {
	do_store(opvalues[0] | opvalues[1]);
    }

    public void op_bitxor() throws GlulxException
    {
	do_store(opvalues[0] ^ opvalues[1]);
    }

    public void op_bitnot() throws GlulxException
    {
	do_store(~opvalues[0]);
    }

    public void op_shiftl() throws GlulxException
    {
	if ((opvalues[1] < 0) || (opvalues[1] >= 32))
	    do_store(0);
	else
	    do_store(opvalues[0]<<opvalues[1]);
    }

    public void op_ushiftr() throws GlulxException
    {
	if ((opvalues[1] < 0) || (opvalues[1] >= 32))
	    do_store(0);
	else
	    do_store(opvalues[0]>>>opvalues[1]);
    }

    public void op_sshiftr() throws GlulxException
    {
	if ((opvalues[1] < 0) || (opvalues[1] >= 32))
	    if (opvalues[0] >= 0)
		do_store(0);
	    else
		do_store(-1);
	else
	    do_store(opvalues[0]>>opvalues[1]);
    }

    public void do_branch(int offset) throws GlulxException
    {
	if ((offset == 0) || (offset == 1))
	    gm.do_return(offset);
	else {
	    gm.pc = gm.pc + offset - 2;
	}
    }

    public void op_jump() throws GlulxException
    {
	do_branch(opvalues[0]);
    }

    public void op_jz() throws GlulxException
    {
	if (opvalues[0] == 0)
	    do_branch(opvalues[1]);
    }

    public void op_jnz() throws GlulxException
    {
	if (opvalues[0] != 0)
	    do_branch(opvalues[1]);
    }

    public void op_jeq() throws GlulxException
    {
	if (opvalues[0] == opvalues[1])
	    do_branch(opvalues[2]);
    }

    public void op_jne() throws GlulxException
    {
	if (opvalues[0] != opvalues[1])
	    do_branch(opvalues[2]);
    }

    public void op_jlt() throws GlulxException
    {
	if (opvalues[0] < opvalues[1])
	    do_branch(opvalues[2]);
    }

    public void op_jle() throws GlulxException
    {
	if (opvalues[0] <= opvalues[1])
	    do_branch(opvalues[2]);
    }

    public void op_jgt() throws GlulxException
    {
	if (opvalues[0] > opvalues[1])
	    do_branch(opvalues[2]);
    }

    public void op_jge() throws GlulxException
    {
	if (opvalues[0] >= opvalues[1])
	    do_branch(opvalues[2]);
    }

    public void op_jltu() throws GlulxException
    {
	if ((opvalues[0]&0xFFFFFFFFL) < (opvalues[1]&0xFFFFFFFFL))
	    do_branch(opvalues[2]);
    }

    public void op_jleu() throws GlulxException
    {
	if ((opvalues[0]&0xFFFFFFFFL) <= (opvalues[1]&0xFFFFFFFFL))
	    do_branch(opvalues[2]);
    }

    public void op_jgtu() throws GlulxException
    {
	if ((opvalues[0]&0xFFFFFFFFL) > (opvalues[1]&0xFFFFFFFFL))
	    do_branch(opvalues[2]);
    }

    public void op_jgeu() throws GlulxException
    {
	if ((opvalues[0]&0xFFFFFFFFL) >= (opvalues[1]&0xFFFFFFFFL))
	    do_branch(opvalues[2]);
    }

    public void op_jumpabs() throws GlulxException
    {
	gm.pc = opvalues[0];
    }

    public void op_getiosys() throws GlulxException
    {
	do_store(gm.current_iosys);
	do_store(gm.current_iosys_rock);
    }

    public void op_setiosys() throws GlulxException
    {
	gm.current_iosys = opvalues[0];
	gm.current_iosys_rock = opvalues[1];
    }
    
    public void op_streamchar() throws GlulxException
    {
	gm.do_streamchar(opvalues[0]);
    }

    public void op_streamnum() throws GlulxException
    {
	gm.do_streamnativestring(Integer.toString(opvalues[0]));
    }

    public void op_streamstr() throws GlulxException
    {
	gm.do_streamstring(opvalues[0]);
    }

    public void op_getmemsize() throws GlulxException
    {
	do_store((int)(gm.ENDMEM + gm.g_memory.getDRAMSize()));
    }

    public void op_setmemsize() throws GlulxException
    {
	long dramsize = (opvalues[0]&0xFFFFFFFFL) - gm.ENDMEM;
	if ((dramsize < 0) || ((dramsize & 0xFF)!=0))
	    do_store(1);
	if (gm.g_memory.setDRAMSize((int)dramsize))
	    do_store(0);
	else
	    do_store(1);
    }
  
    public void op_aload() throws GlulxException
    {
	do_store(gm.g_memory.readWord(opvalues[0] + 4*opvalues[1]));
    }
    
    public void op_aloads() throws GlulxException
    {
	do_store(gm.g_memory.readShort(opvalues[0] + 2*opvalues[1]));
    }
    
    public void op_aloadb() throws GlulxException
    {
	do_store(gm.g_memory.readByte(opvalues[0] + opvalues[1]));
    }

    public void op_astore() throws GlulxException
    {
	gm.g_memory.writeWord(opvalues[0] + 4*opvalues[1], opvalues[2]);
    }
    
    public void op_astores() throws GlulxException
    {
	gm.g_memory.writeShort(opvalues[0] + 2*opvalues[1], opvalues[2]);
    }
    
    public void op_astoreb() throws GlulxException
    {
	gm.g_memory.writeByte(opvalues[0] + opvalues[1], opvalues[2]);
    }
    
    public void op_aloadbit() throws GlulxException
    {
	int bytenum, bitnum, mask;
	if (opvalues[1] < 0) {
	    bytenum = opvalues[0] + ((opvalues[1]+1)/8) - 1;
	    bitnum = (opvalues[1]%8 + 8)%8;
	}
	else {
	    bytenum = opvalues[0] + opvalues[1]/8;
	    bitnum = opvalues[1]%8;
	}
	mask = 1<<bitnum;
	do_store((gm.g_memory.readByte(bytenum&0xFFFFFFFFL)&mask)>>bitnum);
    }

    public void op_astorebit() throws GlulxException
    {
	int bytenum, bitnum, mask;
	if (opvalues[1] < 0) {
	    bytenum = opvalues[0] + ((opvalues[1]+1)/8) - 1;
	    bitnum = (opvalues[1]%8 + 8)%8;
	}
	else {
	    bytenum = opvalues[0] + opvalues[1]/8;
	    bitnum = opvalues[1]%8;
	}
	mask = 1<<bitnum;
	int oldval = gm.g_memory.readByte(bytenum&0xFFFFFFFFL);
	if (opvalues[2] != 0)
	    gm.g_memory.writeByte(bytenum&0xFFFFFFFFL, oldval | mask);
	else
	    gm.g_memory.writeByte(bytenum&0xFFFFFFFFL, oldval & (mask^0xFF));
	    
    }

    public static final int Search_KeyIndirect = 1;
    public static final int Search_ZeroKeyTerminates = 2;
    public static final int Search_ReturnIndex = 4;

    public void op_binarysearch() throws GlulxException
    {
	Log.log(Log.DEBUG, "op_binarysearch", "Entering");
	int keysize = opvalues[1];
	int orig_start = opvalues[2];
	int structsize = opvalues[3];
	int numstructs = opvalues[4];
	int keyoffset = opvalues[5];
	int options = opvalues[6];
	
	int left, right, current;
	int start = orig_start;
	int cmpresult;
	byte [] key = new byte[keysize];
	int i;

	if ((options & Search_KeyIndirect) != 0) {
	    for (i = 0; i < keysize; i++) {
		key[i] = (byte)gm.g_memory.readByte((opvalues[0]&0xFFFFFFFFL) + i);
	    }
	}
	else {
	    for (i = keysize - 1; i >= 0; i--) {
		key[i] = (byte)(opvalues[0]&0xFF);
		opvalues[0] >>>= 8;
	    }
	}
	
	left = 0; right = numstructs - 1;
	while (left <= right) {
	    current = (left + right) / 2;
	    Log.log(Log.DEBUG_EXTREME, "op_binary_search", "l: " + left + " r: " + right + " c: " + current);
	    start = orig_start + current * structsize;
	    cmpresult = 0;
	    for (i = 0;  (i < keysize) && (cmpresult == 0); i++) {
		int cmpval = gm.g_memory.readByte((start&0xFFFFFFFFL) + keyoffset + i);
		Log.log(Log.DEBUG_EXTREME, "op_binarysearch",
			"key["+i+"] = " + key[i] + " cmpval = " + cmpval);
		cmpresult = (key[i]&0xFF) - cmpval;
	    }
	    if (cmpresult == 0) {
		if ((options & Search_ReturnIndex) != 0)
		    do_store(current);
		else
		    do_store(start);
		return;
	    }
	    else if (cmpresult > 0) /* key > cmpval */
		left = current + 1;
	    else
		right = current - 1;
	}
	if ((options & Search_ReturnIndex) != 0)
	    do_store(-1);
	else
	    do_store(0);
    }

    public void op_linearsearch() throws GlulxException
    {
	Log.log(Log.DEBUG, "op_linearsearch", "Entering");

	int keysize = opvalues[1];
	int start = opvalues[2];
	int structsize = opvalues[3];
	int numstructs = opvalues[4];
	int keyoffset = opvalues[5];
	int options = opvalues[6];
	byte [] key = new byte[keysize];
	boolean ismatch, iszero;
	int i;
	if ((options & Search_KeyIndirect) != 0) {
	    for (i = 0; i < keysize; i++) {
		key[i] = (byte)gm.g_memory.readByte((opvalues[0]&0xFFFFFFFFL) + i);
	    }
	}
	else {
	    for (i = keysize - 1; i >= 0; i--) {
		key[i] = (byte)(opvalues[0]&0xFF);
		opvalues[0] >>>= 8;
	    }
	}

	int structindex = 0;
	while ((numstructs == -1) || (structindex < numstructs)) {
	    ismatch = iszero = true;
	    for (i = 0; i < keysize; i++) {
		int cmpval = gm.g_memory.readByte((start&0xFFFFFFFFL) + keyoffset + i);
		Log.log(Log.DEBUG_EXTREME, "op_linearsearch",
			"key["+i+"] = " + key[i] + " cmpval = " + cmpval);
		if (key[i] != cmpval) {
		    ismatch = false;
		    break;
		}
		if (cmpval != 0)
		    iszero = false;

	    }
	    if (ismatch) {
		if ((options & Search_ReturnIndex) != 0)
		    do_store(structindex);
		else
		    do_store(start);
		return;
	    }
	    if (iszero && ((options & Search_ZeroKeyTerminates) != 0))
		break;

	    start += structsize;
	    structindex++;
	}
	if ((options & Search_ReturnIndex) != 0)
	    do_store(-1);
	else
	    do_store(0);
	return;
    }

    public void op_linkedsearch() throws GlulxException
    {
	int keysize = opvalues[1];
	int start = opvalues[2];
	int keyoffset = opvalues[3];
	int nextoffset = opvalues[4];
	int options = opvalues[5];
	byte [] key = new byte[keysize];
	boolean ismatch, iszero;
	int i;
	if ((options & Search_KeyIndirect) != 0) {
	    for (i = 0; i < keysize; i++) {
		key[i] = (byte)gm.g_memory.readByte((opvalues[0]&0xFFFFFFFFL) + i);
	    }
	}
	else {
	    for (i = keysize - 1; i >= 0; i--) {
		key[i] = (byte)(opvalues[0]&0xFF);
		opvalues[0] >>>= 8;
	    }
	}

	int structindex = 0;
	while (true) {
	    ismatch = iszero = true;;
	    for (i = 0; i < keysize; i++) {
		int cmpval = gm.g_memory.readByte((start&0xFFFFFFFFL) + keyoffset + i);
		if (key[i] != cmpval) {
		    ismatch = false;
		    break;
		}
		if (key[i] != 0)
		    iszero = false;

	    }
	    if (ismatch) {
		do_store(start);
		return;
	    }
	    if (iszero && ((options & Search_ZeroKeyTerminates) != 0))
		break;

	    start = gm.g_memory.readWord((start&0xFFFFFFFFL) + nextoffset);
	    structindex++;
	}
	do_store(0);
	return;
    }

    protected int stackcount()
    {
	return gm.g_stack.getSP() - gm.fp - gm.framelen;
    }

    public void op_stkcount() throws GlulxException
    {
	do_store(stackcount());
    }

    public void op_stkpeek() throws GlulxException
    {
	if (opvalues[0] >= stackcount())
	    throw new GlulxException("Tried to peek beyond frame");
	do_store(gm.g_stack.peekWord(opvalues[0]));
    }

    public void op_stkswap() throws GlulxException
    {
	int tmp;
	if (stackcount() < 2)
	    throw new GlulxException("Tried to swap stack with too few values");
	tmp = gm.g_stack.peekWord(0);
	gm.g_stack.pokeWord(0, gm.g_stack.peekWord(1));
	gm.g_stack.pokeWord(1, tmp);
    }

    public void op_stkcopy() throws GlulxException
    {
	if (stackcount() < opvalues[0])
	    throw new GlulxException("Tried to copy too many values from stack");
	if (Log.willLog(Log.DEBUG_EXTREME, "op_stkcopy")) {
	    String logme = "";
	    for (int i = opvalues[0]-1; i >= 0; i--) {
		logme += Integer.toHexString(gm.g_stack.peekWord(i)) + " ";
	    }
	    Log.log(Log.DEBUG_EXTREME, "op_stkcopy", logme);
	}
	
	for (int i = 0; i < opvalues[0]; i++)
	    gm.g_stack.pushWord(gm.g_stack.peekWord(opvalues[0]-1));

	if (Log.willLog(Log.DEBUG_EXTREME, "op_stkcopy")) {
	    String logme = "";
	    for (int i = (opvalues[0]*2)-1; i >= 0; i--) {
		logme += Integer.toHexString(gm.g_stack.peekWord(i)) + " ";
	    }
	    Log.log(Log.DEBUG_EXTREME, "op_stkcopy", logme);
	}
    }

    public void op_stkroll() throws GlulxException
    {
	int tmp;
	int rolldepth = opvalues[0];
	if (stackcount() < opvalues[0])
	    throw new GlulxException("Tried to roll stack with too few values");
	int rollcount = opvalues[1] % rolldepth;
	if (rollcount < 0)
	    rollcount = rollcount + rolldepth;
	//	    	for (int j = 0; j < rollcount; j++) {
	//		    tmp = gm.g_stack.peekWord(0);
	//		    for (int i = 0; i < rolldepth - 1; i++) {
	//			gm.g_stack.pokeWord(i, gm.g_stack.peekWord(i + 1));
	//		    }
	//		    gm.g_stack.pokeWord(rolldepth-1, tmp);
	//		}
	//		/* inefficient as heck */

	// The algorithm below is weird, but it seems to be fairly fast
	// and it uses constant storage
	// If rollcount is relatively prime to rolldepth, the while loop
	// executes once.  The i and nexti values take on all possible
	// values between 0 and rollcount-1 inclusive.  (by some number 
	// theory theorem I've forgotten the name of).  Otherwise, the
	// i and nexti values describe smaller cycle within that range.
	// The offset is then used to move to another cycle.  This should
	// cover all values, though I can't guarantee it because I've
	// forgotten too much number theory
	int totroll = 0;
	int offset = 0;
	int i, nexti;
	while (totroll < rollcount) {
	    tmp = gm.g_stack.peekWord(offset);
	    i = offset;
	    nexti = (i + rollcount) % rolldepth;
	    while (nexti != offset) {
		gm.g_stack.pokeWord(i, gm.g_stack.peekWord(nexti));
		totroll++;
		i = nexti;
		nexti = (i + rollcount) % rolldepth;
	    }
	    gm.g_stack.pokeWord(i, tmp);
	    totroll++;
	    offset++;
	}
    }

    public void op_gestalt() throws GlulxException
    {
	do_store(gm.gestalt(opvalues[0], opvalues[1]));
    }

    protected Object get_glk_arg(int basic_type, int glulx_arg) 
	throws GlulxException 
    {
	switch (basic_type) {
	case GlkArgument.type_Is:
	case GlkArgument.type_Iu:
	    return new Integer(glulx_arg);
	case GlkArgument.type_Cs:
	case GlkArgument.type_Cu:
	case GlkArgument.type_Cn:
	    return new Byte((byte)glulx_arg);
	default:
	    if ((basic_type >= GlkArgument.type_Qa) &&
		(basic_type <= GlkArgument.type_Qz)) {
		return gm.toGlkOpaqueRef(glulx_arg,
					 basic_type - GlkArgument.type_Qa);
	    }
	    throw new GlulxException("Glk Type Not Supported " + basic_type);
	}
    }

    protected int get_glk_out_arg(int basic_type, Object glk_arg) 
	throws GlulxException 
    {
	switch (basic_type) {
	case GlkArgument.type_Is:
	case GlkArgument.type_Iu:
	    return ((Integer)glk_arg).intValue();
	case GlkArgument.type_Cs:
	    return ((Byte)glk_arg).intValue();
	case GlkArgument.type_Cu:
	case GlkArgument.type_Cn:
	    return ((Byte)glk_arg).intValue()&0xFF;
	default:
	    if ((basic_type >= GlkArgument.type_Qa) &&
		(basic_type <= GlkArgument.type_Qz)) {
		return gm.toGlulxOpaqueRef(glk_arg,
					 basic_type - GlkArgument.type_Qa);
	    }
	    throw new GlulxException("Glk Type Not Supported " + basic_type);
	}
    }

    public void op_glk() throws GlulxException
    {
	int selector = opvalues[0];
	int argspassed = opvalues[1];
	int rawargs[] = new int[argspassed];
	int retval = 0;

	Log.log(Log.DEBUG, "op_glk", "Glk: " + selector + " nargs: " + argspassed);
	if (Log.willLog(Log.DEBUG, "op_glk")) {
	    for (int i = 0; i < argspassed; i++) {
		rawargs[i] = gm.g_stack.popWord();
		Log.log(Log.DEBUG, "op_glk", Integer.toString(rawargs[i]));
	    }
	}

	int rawargspos = 0;

	GlkArgument [] argtypes = GlkArgument.CreateArgsFromString(gm.glkdis.gidispatch_prototype(selector));
	int nglkargs = argtypes.length;
	Object args[] = new Object[nglkargs];

	for (int i = 0; i < nglkargs; i++) {
	    if (argtypes[i].returnval) {
		Log.log(Log.DEBUG, "op_glk", i + " returnval");
		break;
	    }
	    int glulx_arg = rawargs[rawargspos++];
	    int basic_type = argtypes[i].basic_type;
	    Log.log(Log.DEBUG, "op_glk", i + " " + argtypes[i].isref);
	    if (!argtypes[i].isref) {
		args[i] = get_glk_arg(basic_type, glulx_arg);
	    }
	    else if (argtypes[i].isarray) {
		int arraylen = rawargs[rawargspos++];
		if (glulx_arg == 0) {
		    args[i] = null;
		    if (argtypes[i].required)
			throw new GlulxException("Illegal null passed to Glk");
		    break;
		}
		switch (basic_type) {
		case GlkArgument.type_Cu:
		case GlkArgument.type_Cs:
		case GlkArgument.type_Cn:
		    byte ba[] = new byte[arraylen];
		    int pos = glulx_arg;
		    if (argtypes[i].copy_in) {
			for (int j = 0; j < arraylen; j++) {
			    ba[j] = (byte)gm.g_memory.readByte(pos++);
			}
		    }
		    args[i] = ba;
		    break;
		case GlkArgument.type_Iu:
		case GlkArgument.type_Is:
		    int ia[] = new int[arraylen];
		    pos = glulx_arg;
		    if (argtypes[i].copy_in) {
			for (int j = 0; j < arraylen; j++) {
			    ia[j] = gm.g_memory.readWord(pos);
			    pos += 4;
			}
		    }
		    args[i] = ia;
		    break;
		default:
		    throw new GlulxException("Illegal basic type for array" + basic_type);
		}
		if (argtypes[i].retained)
		    gm.glkreg.expect_array(args[i], glulx_arg);
	    }
	    else {
		/* Regular or struct reference */
		if (glulx_arg == 0)  {
		    if (argtypes[i].required)
			throw new GlulxException("Illegal null passed to Glk");
		    args[i] = null;
		    break;
		}

		if (argtypes[i].structure) {
		    /* doesn't handle non-basic-type-containing structs */
		    int nstructargs = argtypes[i].structargs.length;
		    Object [] structargs = new Object[nstructargs];
		    args[i] = structargs;
		    for (int j = 0; j < nstructargs; j++) {
			if (!argtypes[i].copy_in)
			    structargs[j] = GlkDispatcher.getNoChangeReference();
			else if (glulx_arg == -1)
			    structargs[j] = get_glk_arg(argtypes[i].structargs[j].basic_type,
							gm.g_stack.popWord());
			else {
			    structargs[j] = get_glk_arg(argtypes[i].structargs[j].basic_type,
							gm.g_memory.readWord(glulx_arg));
			    glulx_arg += 4;
			}
		    }
		}
		else if (argtypes[i].copy_in) { /* normal reference */
		    int in_val;
		    if (glulx_arg == -1)
			in_val = gm.g_stack.popWord();
		    else
			in_val = gm.g_memory.readWord(glulx_arg);
		    args[i] = get_glk_arg(basic_type, in_val);
		}
		else { /* normal but noncopied reference */
		    args[i] = GlkDispatcher.getNoChangeReference();
		}
	    }
	}
	if (Log.willLog(Log.DEBUG, "op_glk")) {
	    for (int i = 0; i < nglkargs; i++) {
		if (args[i] != null)
		    Log.log(Log.DEBUG, "op_glk", 
			    args[i].getClass().getName() + " " + args[i]);
		else
		    Log.log(Log.DEBUG, "op_glk", "(null)");
	    }
	}
	gm.glkdis.gidispatch_call(selector, nglkargs, args);
	if (Log.willLog(Log.DEBUG, "op_glk")) {
	    for (int i = 0; i < nglkargs; i++) {
		if (args[i] != null)
		    Log.log(Log.DEBUG, "op_glk", 
			    args[i].getClass().getName() + " " + args[i]);
		else
		    Log.log(Log.DEBUG, "op_glk", "(null)");
	    }
	}
	rawargspos = 0;
	for (int i = 0; i < nglkargs; i++) {
	    int basic_type = argtypes[i].basic_type;
	    int glulx_arg = 0;
	    if (argtypes[i].isref && argtypes[i].copy_out && 
		(args[i] != GlkDispatcher.getNoChangeReference()) &&
		!argtypes[i].retained &&
		(argtypes[i].returnval ||
		((glulx_arg = rawargs[rawargspos]) != 0))) {
		if (argtypes[i].isarray) {
		    int arraylen = rawargs[rawargspos+1];
		    switch (basic_type) {
		    case GlkArgument.type_Cu:
		    case GlkArgument.type_Cs:
		    case GlkArgument.type_Cn:
			byte ba[] = (byte [])args[i];
			for (int j = 0; j < arraylen; j++) {
			    gm.g_memory.writeByte(glulx_arg++, ba[j]);
			}
			break;
		    case GlkArgument.type_Iu:
		    case GlkArgument.type_Is:
			int ia[] = (int [])args[i];
			for (int j = 0; j < arraylen; j++) {
			    gm.g_memory.writeWord(glulx_arg, ia[j]);
			    glulx_arg += 4;
			}
			break;
		    default:
			throw new GlulxException("Illegal basic type for array" + basic_type);
		    }
		}
		else if (argtypes[i].structure) {
		  /* only works for simple structures */

		  int len = argtypes[i].structargs.length;
		  Object []structargs = (Object [])args[i];
		  for (int j = 0; j < len; j++) {
		    if (glulx_arg == -1)
		      gm.g_stack.pushWord(
			    get_glk_out_arg(argtypes[i].structargs[j].basic_type,
			    structargs[j]));
		    else {
		      gm.g_memory.writeWord(glulx_arg,
			    get_glk_out_arg(argtypes[i].structargs[j].basic_type,
			    structargs[j]));
		      glulx_arg += 4;
		    }
		  }
		}
		else {
		    int outval = get_glk_out_arg(basic_type, args[i]);
		    if (argtypes[i].returnval)
			retval = outval;
		    else if (glulx_arg == -1)
			gm.g_stack.pushWord(outval);
		    else
			gm.g_memory.writeWord(glulx_arg, outval);
		}
	    }
	    /* if it's not a reference or not copy-out, do nothing */
	    rawargspos++;
	    if (argtypes[i].isarray)
		rawargspos++;
	}
	do_store(retval);
    }

    public void op_quit() throws GlulxException
    {
	gm.quit();
    }
    
    public void op_random() throws GlulxException
    {
	if (opvalues[0] == 0)
	    do_store(gm.rng.nextInt());
	else
	    do_store(gm.rng.nextInt(opvalues[0]));
    }

    public void op_setrandom() throws GlulxException
    {
	if (opvalues[0] == 0)
	    gm.rng.setSeed(System.currentTimeMillis());
	else
	    gm.rng.setSeed(opvalues[0]);
    }
    
    public void op_saveundo() throws GlulxException
    {
	do_store(1); /* fails always */
    }

    public void op_restoreundo() throws GlulxException
    {
	do_store(1); /* always fails */
    }
}

