package russotto.jglulx;
import java.io.DataInputStream;
import java.io.IOException;
public class GMemory
{
    byte ROM[]; 
    byte RAM[];
    byte dRAM[];
    int RAMStart;
    int dRAMStart;
    
    public GMemory(long romsize, long ramsize, long dRAMsize)
    {
	ROM = new byte[(int)romsize];
	RAM = new byte[(int)ramsize];
	if (dRAMsize != 0)
	    dRAM = new byte[(int)dRAMsize];
	// none can be > 2GB.  Oh well
	RAMStart = ROM.length;
	dRAMStart = ROM.length + RAM.length;
    }

    public GMemory(long romsize, long ramsize) 
    {
	this(romsize, ramsize, 0);
    }
    
    public long getRAMSize() {
	return RAM.length;
    }

    public long getROMSize() {
	return ROM.length;
    }

    public long getDRAMSize() {
	if (dRAM != null)
	    return dRAM.length;
	return 0;
    }

    public boolean setDRAMSize(int drsize) {
	byte []ndram = new byte[drsize];
	if (ndram == null)
	    return false;
	if (dRAM != null) {
	    int length = drsize;
	    
	    if (length > dRAM.length)
		length = dRAM.length;
	    
	    System.arraycopy(dRAM, 0, ndram, 0, length);
	}
	dRAM = ndram;
	return true;
    }

    public void load(DataInputStream in, long start, long length) throws IOException {
	if ((start < RAMStart) && (length > 0)) {
	    // must be some ROM
	    long romlength = RAMStart - start;
	    if (length < romlength) romlength = length;
	    in.readFully(ROM, (int)start, (int)romlength);
	    length -= romlength;
	    start += romlength;
	}
	if ((start < dRAMStart) && (length > 0)) {
	    // must be some RAM
	    long ramlength = dRAMStart - start;
	    if (length < ramlength) ramlength = length;
	    in.readFully(RAM, (int)(start - RAMStart), (int)ramlength);
	    length -= ramlength;
	    start += ramlength;
	}
	if (length > 0) {
	    in.readFully(dRAM, (int)(start - dRAMStart), (int)length);
	}
    }

    void copyOut(long src_position, byte[] dst, int dst_position, int length)
    {
	if ((src_position < RAMStart) && (length > 0)) {
	    long romlength = RAMStart - src_position;
	    if (length < romlength) romlength = length;
	    System.arraycopy(ROM, (int)src_position, dst, dst_position, 
			     (int)romlength);
	    src_position += romlength;
	    dst_position += (int)romlength;
	    length -= (int)romlength;
	}

	if ((src_position < dRAMStart) && (length > 0)) {
	    long ramlength = dRAMStart - src_position;
	    if (length < ramlength) ramlength = length;
	    System.arraycopy(RAM, (int)src_position, dst, dst_position, 
			     (int)ramlength);
	    src_position += ramlength;
	    dst_position += (int)ramlength;
	    length -= (int)ramlength;
	}
	
	if (length > 0)
	    System.arraycopy(dRAM, (int)src_position, dst, dst_position, length);
    }

    /*
      public int readRAMByte(long addr) {
	return readByte(addr - RAMStart);
    }

    public int readRAMShort(long addr) {
	return readShort(addr - RAMStart);
    }

    public int readRAMWord(long addr) {
	return readWord(addr - RAMStart);
    }

    public void writeRAMByte(long addr, int val) {
	writeByte(addr - RAMStart, val);
    }

    public void writeRAMShort(long addr, int val) {
	writeShort(addr - RAMStart, val);
    }

    public void writeRAMWord(long addr, int val) {
	writeWord(addr - RAMStart, val);
    }
    */

    public int readByte(long addr) {
	if (addr < RAMStart)
	    return ROM[(int)addr] & 0xFF;
	else if (addr < dRAMStart)
	    return RAM[(int)(addr - RAMStart)] & 0xFF;
	else
	    return dRAM[(int)(addr - dRAMStart)] & 0xFF;
    }

    public int readShort(long addr) {
	return (readByte(addr)<<8) | readByte(addr+1);
    }

    public int readWord(long addr) {
	return (readByte(addr) << 24) |
	    (readByte(addr+1) << 16) |
	    (readByte(addr+2) << 8)  |
	    readByte(addr+3);
    }

    void writeROMByte(long addr, int val) {
	ROM[(int)addr] = (byte)val;
    }

    void writeROMShort(long addr, int val) {
	ROM[(int)addr] = (byte)(val>>8);
	ROM[(int)addr+1] = (byte)(val&0xFF);
    }

    void writeROMWord(long addr, long val)  // convenience call
    {
	writeROMWord(addr, (int)val);
    }

    void writeROMWord(long addr, int val) {
	ROM[(int)addr] = (byte)(val>>>24);
	ROM[(int)addr+1] = (byte)((val>>>16)&0xFF);
	ROM[(int)addr+2] = (byte)((val>>>8)&0xFF);
	ROM[(int)addr+3] = (byte)(val&0xFF);
    }

    public void writeByte(long addr, int val) {
	if (addr < RAMStart)
	    System.err.println("Tried to write to ROM at " + 
			       Long.toHexString(addr) + " value " +
			       Integer.toHexString(val));
	else if (addr < dRAMStart)
	    RAM[(int)(addr - RAMStart)] = (byte)val;
	else
	    dRAM[(int)(addr - dRAMStart)] = (byte)val;
    }

    public void writeShort(long addr, int val) {
	writeByte(addr++, (val>>>8));
	writeByte(addr, val&0xFF);
    }

    public void writeWord(long addr, int val) {
	writeByte(addr++, (val>>>24));
	writeByte(addr++, (val>>>16)&0xFF);
	writeByte(addr++, (val>>>8)&0xFF);
	writeByte(addr, val&0xFF);
    }
    
    public int checksum() {
	int i;
	int checksum = 0;
	int length = dRAMStart;
	if (dRAM != null)
	    length += dRAM.length;

	for (i = 0; i < length; i += 4) {
	    checksum += readWord(i);
	}
	return checksum;
    }
}
