package russotto.jglulx;
import glk.Glk;
import glk.GlkExitException;
import glk.GlkProgram;
import glk.GlkProxy;
import glk.GlkCLStartup;
import glk.GlkDispatcher;
import glk.GlkDispatch;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.Random;
public class Glulx extends GlkProxy implements GlkProgram, GlkCLStartup
{
    public final static int iosys_Null = 0;
    public final static int iosys_Filter = 1;
    public final static int iosys_Glk = 2;
    public final static int Glulx_MajorVersion = 0x0002;
    public final static int Glulx_MinorVersion = 0x0000;
    public final static int Glulx_Magic =  0x476C756C; /* 'Glul' */

    GStack g_stack;
    GMemory g_memory;
    GlkDispatcher glkdis;
    GlulxGlkRegistry glkreg;
    Random rng;
    int pc;
    int fp;
    int current_iosys;
    int current_iosys_rock;
    byte initialRAM[];
    long RAMSTART, EXTSTART, ENDMEM;
    int glulx_minor;
    int glulx_veryminor;
    int stored_checksum;
    long max_stack;
    int framelen; 
    int localspos;
    int string_decoding_tbl;

    InputStream storystream;
  
    public Glulx() {
    }

    Glulx(InputStream instream) throws GlulxException, IOException
    {
	initialize(instream);
    }

    public void glk_commandline_startup(Glk glk, String argv[]) {
	try {
	    storystream = new FileInputStream(argv[0]);
	}
	catch (Exception e) {
	    e.printStackTrace();
	    glk.glk_exit();
	}
    }

    public void glk_main(Glk glk) 
    {
	this.glk = glk;
	this.glkdis = new GlkDispatcher(glk);
	this.glkreg = new GlulxGlkRegistry(this, (GlkDispatch)glk);
	this.rng =  new Random();
	Log.setConsoleLevel(Log.WARNING);
	Log.setFileLevel(Log.DEBUG);
	try {
	    initialize(storystream);
	    testrun();
	}
	catch (GlkExitException glke) {}
	catch (Exception e) {
	    Log.flush();
	    e.printStackTrace();
	    System.err.println("pc = " + Integer.toHexString(pc));
	    System.err.println("fp = " + Integer.toHexString(fp));
	    if (g_stack != null)
		System.err.println("sp = " + Integer.toHexString(g_stack.getSP()));
	    if (e instanceof RuntimeException)
		throw (RuntimeException)e;
	}
    }

    protected void initialize(InputStream instream) throws GlulxException, IOException
    {
	DataInputStream glulxstream = new DataInputStream(instream);
	int magic = glulxstream.readInt();
	if (magic != Glulx_Magic)
	    throw new GlulxException("Magic number incorrect " + Integer.toHexString(magic));
	int major = glulxstream.readUnsignedShort();
	if (major != Glulx_MajorVersion)
	    throw new GlulxException("Glulx Version incorrect " + major);
	glulx_minor = glulxstream.readUnsignedByte();
	glulx_veryminor = glulxstream.readUnsignedByte();
	RAMSTART = glulxstream.readInt() & 0xFFFFFFFF;
	EXTSTART = glulxstream.readInt() & 0xFFFFFFFF;
	ENDMEM = glulxstream.readInt() & 0xFFFFFFFF;
	g_memory = new GMemory(RAMSTART, ENDMEM - RAMSTART);
	g_memory.writeROMWord(0, Glulx_Magic);
	g_memory.writeROMShort(4, Glulx_MajorVersion);
	g_memory.writeROMByte(6, glulx_minor);
	g_memory.writeROMByte(7, glulx_veryminor);
	g_memory.writeROMWord(8, RAMSTART);
	g_memory.writeROMWord(12, EXTSTART);
	g_memory.writeROMWord(16, ENDMEM);
	g_memory.load(glulxstream, 20, EXTSTART-20);
	string_decoding_tbl = g_memory.readWord(28);
	stored_checksum = g_memory.readWord(32);
	int calc_checksum = g_memory.checksum();
	if ((g_memory.checksum() - stored_checksum) != stored_checksum) {
	    /* most checksummers would store the twos-complement of
	       the checksum calculated with 0 as the CS value in the
	       checksum location.
	       Glk doesn't, hence the odd calculation */
	    throw new GlulxException("Checksum bad " + 
				     Integer.toHexString(calc_checksum) + 
				     " should be " + 
				     Integer.toHexString(stored_checksum));
	}
	max_stack = g_memory.readWord(20) & 0xFFFFFFFF;
	diagstart();
	g_stack = new GStack(max_stack);
	pc = g_memory.readWord(24);
	make_callstack(null);
    }

    public void testrun() throws GlulxException 
    {
	GInstruction ins = new GInstruction(this);
	while (true) {
	    ins.decode_instruction();
	    ins.log_instruction(pc);
	    ins.execute();
	}
    }

    public int getCodeByte() 
    {
	//	int thebyte = g_memory.readByte(pc++);
	//	System.err.println(Integer.toHexString(thebyte));
	//	return thebyte;
	return g_memory.readByte(pc++);
	
    }

    public int getCodeShort() 
    {
	int retval = g_memory.readShort(pc);
	pc += 2;
	return retval;
    }

    public int getCodeWord() 
    {
	int retval = g_memory.readWord(pc);
	pc += 4;
	return retval;
    }
    
    public int readLocalByte(long addr)
	throws GlulxException
    {
	if (addr >= framelen)
	    throw new GlulxException("Tried to access non-local local");
	return g_stack.peekByteAbs(fp + localspos + (int)addr);
    }

    public int readLocalShort(long addr)
	throws GlulxException
    {
	if (addr >= framelen - 1)
	    throw new GlulxException("Tried to access non-local local");
	return g_stack.peekShortAbs(fp + localspos + (int)addr);
    }

    public int readLocalWord(long addr)
	throws GlulxException
    {
	if (addr >= framelen - 3)
	    throw new GlulxException("Tried to access non-local local");
	return g_stack.peekWordAbs(fp + localspos + (int)addr);
    }

    public void writeLocalByte(long addr, int val)
	throws GlulxException
    {
	if (addr >= framelen)
	    throw new GlulxException("Tried to access non-local local");
	g_stack.pokeByteAbs(fp + localspos + (int)addr, val);
    }

    public void writeLocalShort(long addr, int val)
	throws GlulxException
    {
	if (addr >= framelen - 1)
	    throw new GlulxException("Tried to access non-local local");
	g_stack.pokeShortAbs(fp + localspos + (int)addr, val);
    }

    public void writeLocalWord(long addr, int val)
	throws GlulxException
    {
	if (addr >= framelen - 3)
	    throw new GlulxException("Tried to access non-local local " + Long.toHexString(addr));
	g_stack.pokeWordAbs(fp + localspos + (int)addr, val);
    }

    protected void diagstart() {
	System.err.println("Loaded Glulx file");
	System.err.println("Glulx Version " + Glulx_MajorVersion + "." +
			   glulx_minor + "." + glulx_veryminor);
	System.err.println("RAMSTART: " + Long.toHexString(RAMSTART));
	System.err.println("EXTSTART: " + Long.toHexString(EXTSTART));
	System.err.println("ENDMEM: " + Long.toHexString(ENDMEM));
	System.err.println("Stack size: " + Long.toHexString(max_stack));
	System.err.println("Checksum(stored): " + 
			   Long.toHexString(0xFFFFFFFFL & stored_checksum));
	System.err.println("Checksum(calculated): " + 
			   Long.toHexString(0xFFFFFFFFL & g_memory.checksum()));
	System.err.println("Checksum(calculated - stored): " + 
			   Long.toHexString(0xFFFFFFFFL & (g_memory.checksum() - stored_checksum)));
    }

    public void make_callframe(int desttype, int destaddr)
    {
	g_stack.pushWord(desttype);
	g_stack.pushWord(destaddr);
	g_stack.pushWord(pc);
	g_stack.pushWord(fp);
    }

    public void make_callframe(int desttype)
    /* for type-11 callframes */
    {
	g_stack.pushWord(desttype);
	g_stack.pushWord(0);
	g_stack.pushWord(pc);
	g_stack.pushWord(fp);
    }

    public void make_callframe(int desttype, int arg1, int arg2)
    /* for type-10 callframes 
       arg1 = byte;
       arg2 = bit;

       for type-12 callframes
       arg1 = integer;
       arg2 = position;
       
       for type-13 callframes
       arg1 = position;
       arg2 = 0;
     */
     
    {
	g_stack.pushWord(desttype);
	g_stack.pushWord(arg2); /* "destaddr" */
	g_stack.pushWord(arg1); /* "pc" */
	g_stack.pushWord(fp);
    }

    public void make_callstack(int args[]) 
    throws GlulxException
    {

	int funtype;
	fp = g_stack.getSP();
	g_stack.pushWord(0); /* placeholder for frame length */
	g_stack.pushWord(0); /* placeholder for locals pos */
	funtype = getCodeByte();
	if ((funtype != 0xC0) && (funtype != 0xC1))
	    throw new GlulxException("Bad Function Type " + Integer.toHexString(funtype));
	int localtype;
	int localcount;
	int localbytes = 0;
	int format_ptr = g_stack.getSP();
	int args_ptr;
	do {
	    localtype = getCodeByte();
	    localcount = getCodeByte();
	    localbytes += (localtype * localcount);
	    g_stack.pushByte(localtype);
	    g_stack.pushByte(localcount);
	}
	while ((localtype != 0) || (localcount != 0));
	g_stack.pad4();
	localspos = g_stack.getSP() - fp;
	switch (funtype) {
	case 0xC0: /* ARGS ON STACK */
	    while (localbytes-- > 0)
		g_stack.pushByte(0);
	    break;
	case 0xC1: /* ARGS IN LOCALS */
	    args_ptr = 0;
	    do {
		localtype =  g_stack.peekByteAbs(format_ptr++);
		localcount = g_stack.peekByteAbs(format_ptr++);
		//		System.err.println("lc = " + localcount);
		//		System.err.println("lt = " + localtype);
		int localval;
		for (int i = 0; i < localcount; i++) {
		    if ((args != null) && (args_ptr < args.length))
			localval = args[args_ptr++];
		    else
			localval = 0;
		    switch (localtype) {
		    case 4:
			g_stack.pushWord(localval);
			break;
		    case 2:
			g_stack.pushShort(localval&0xFFFF);
			break;
		    case 1:
			g_stack.pushByte(localval&0xFF);
		    }
		}
	    }
	    while ((localtype != 0) || (localcount != 0));
	}
	g_stack.pad4();
	framelen = g_stack.getSP() - fp;
	g_stack.pokeWordAbs(fp, framelen);
	g_stack.pokeWordAbs(fp+4, localspos);
	if ((args != null) && (funtype == 0xC0)) {
	    args_ptr = args.length;
	    while (args_ptr-- > 0) {
		g_stack.pushWord(args[args_ptr]);
	    }
	    g_stack.pushWord(args.length);
	}
	else if (funtype == 0xC0)
	    g_stack.pushWord(0);
    }

    public void do_store(int mode, int addr, int width, int val)
    throws GlulxException
    {
	if (mode != GInstruction.OPMODE_0)
	    Log.log(Log.DEBUG_EXTREME, "do_store", "Storing " + Integer.toHexString(val));
	switch(mode) {
	case GInstruction.OPMODE_SHORT_RAM:
	case GInstruction.OPMODE_BYTE_RAM:
	case GInstruction.OPMODE_WORD_RAM:
	case GInstruction.OPMODE_SHORT_ADDR:
	case GInstruction.OPMODE_BYTE_ADDR:
	case GInstruction.OPMODE_WORD_ADDR:
	    switch (width) {
	    case 4:
		g_memory.writeWord(addr&0xFFFFFFFFL, val);
		break;
	    case 2:
		g_memory.writeShort(addr&0xFFFFFFFFL, val);
		break;
	    case 1:
		g_memory.writeByte(addr&0xFFFFFFFFL, val);
		break;
	    }
	    break;

	case GInstruction.OPMODE_WORD_LOCAL:
	case GInstruction.OPMODE_BYTE_LOCAL:
	case GInstruction.OPMODE_SHORT_LOCAL:
	    switch (width) {
	    case 4:
		writeLocalWord(addr, val);
		break;
	    case 2:
		writeLocalShort(addr, val);
		break;
	    case 1:
		writeLocalByte(addr, val);
		break;
	    }
	    break;

	case GInstruction.OPMODE_0:
	    break;

	case GInstruction.OPMODE_STACK:
	    g_stack.pushWord(val);
	    break;
		
	}
    }

    public void do_return(int value) throws GlulxException
    {
	int mode, addr;
	g_stack.setSP(fp);
	fp = g_stack.popWord();
	pc = g_stack.popWord();
	addr = g_stack.popWord();
	int frametype = g_stack.popWord();
	framelen = g_stack.peekWordAbs(fp);
	localspos = g_stack.peekWordAbs(fp+4);
	if (frametype < 0x10) { /* normal function return */
	    mode = GInstruction.getModeFromDestType(frametype);
	    do_store(mode, addr, 4, value);
	}
	else if (frametype == 0x10) {
	    do_streamE1string(pc, addr, true);
	}
	else if (frametype == 0x13) {
	    do_streamE0string(pc, true);
	}
	else {
	    throw new GlulxException("Unrecognized frame type " + 
				     Integer.toHexString(frametype));
	}
    }

    protected void do_filterchar(int thechar) {
	throw new UnsupportedOperationException();
    }

    public void do_streamchar(int thechar) throws GlulxException {
	switch (current_iosys) {
	case iosys_Null:
	    break;
	case iosys_Filter:
	    make_callframe(0, 0);
	    int args[] = new int[1];
	    args[0] = thechar;
	    pc = current_iosys_rock;
	    make_callstack(args);
	    break;
	case iosys_Glk:
	    glk_put_char((char)thechar);
	    break;
	}
    }
    
    public void do_streamnativestring(String s) {
	switch (current_iosys) {
	case iosys_Null:
	    break;
	case iosys_Filter:
	    for (int i = 0; i < s.length(); i++)
		do_filterchar((int)s.charAt(i));
	    break;
	case iosys_Glk:
	    glk_put_string(s);
	}
    }

    public void do_streamE0string(int straddr) throws GlulxException
    {
	do_streamE0string(straddr, false);
    }

    public void do_streamE0string(int straddr, boolean pushedtype11) throws GlulxException
    {
	int ch;
	switch (current_iosys) {
	case iosys_Null:
	    break;
	case iosys_Filter:
	    ch = g_memory.readByte(straddr++);
	    if (ch == 0) {
		if (pushedtype11) {
		    g_stack.popWord(); /* frame pointer */
		    pc = g_stack.popWord(); /* PC = bitnum for type 10 */
		    straddr = g_stack.popWord(); /* Destaddr */
		    int poptype = g_stack.popWord(); /* 10 or 11 */
		    if (poptype == 0x10)
			do_streamE1string(straddr, pc, true);
		}
		break;
	    }
	    if (!pushedtype11)
		make_callframe(0x11);
	    make_callframe(0x13, straddr, 0);
	    int args[] = new int[1];
	    args[0] = ch;
	    pc = current_iosys_rock;
	    make_callstack(args);
	    break;
	case iosys_Glk:
	    String s = "";
	    ch = g_memory.readByte(straddr++);
	    while (ch != 0) {
		s = s + (char)ch;
		ch = g_memory.readByte(straddr++);
	    }
	    glk_put_string(s);
	    break;
	default:
	    throw new GlulxException("Illegal IO system selected");
	}
    }

    void do_streamE1string(int straddr)	throws GlulxException {
	do_streamE1string(straddr, 0, false);
    }
    
    void do_streamE1string(int straddr, int bitnum, boolean pushedtype11) 
	throws GlulxException
    {

	int tbl = string_decoding_tbl;
	int tbllen = g_memory.readWord(tbl);
	int endtbl = tbl + tbllen;
	int nnodes = g_memory.readWord(tbl+4);
	int rootaddr = g_memory.readWord(tbl + 8);
	int nodeaddr = rootaddr;
	int nodetype;
	boolean stopdecoding = false;

	if ((current_iosys == iosys_Filter) && !pushedtype11) {
	    make_callframe(0x11);
	    pushedtype11 = true;
	}

	while (!stopdecoding) {
	    if ((nodeaddr < tbl) || (nodeaddr > endtbl))
		throw new GlulxException("Failed while decoding Huffman string");
	    nodetype = g_memory.readByte(nodeaddr);
	    switch (nodetype) {
	    case 1:
		if (pushedtype11) {
		    g_stack.popWord(); /* frame pointer */
		    bitnum = pc = g_stack.popWord(); /* PC */
		    straddr = g_stack.popWord(); /* Destaddr */
		    int poptype = g_stack.popWord(); /* 10 or 11 */
		    stopdecoding = (poptype != 0x10);
		}
		else 
		    stopdecoding = true;
		break;
	    case 2:
		if (current_iosys == iosys_Filter) {
		    make_callframe(0x10, straddr, bitnum);
		    stopdecoding = true;
		    int args[] = new int[1];
		    args[0] = g_memory.readByte(nodeaddr+1);
		    pc = current_iosys_rock;
		    make_callstack(args);
		}
		else
		    do_streamchar(g_memory.readByte(nodeaddr+1));
		nodeaddr = rootaddr;
		break;
	    case 3:
		if (current_iosys == iosys_Filter) {
		    make_callframe(0x10, straddr, bitnum);
		    stopdecoding = true;
		}
		do_streamE0string(nodeaddr+1, pushedtype11);
		nodeaddr = rootaddr;
		break;
	    case 0xA:
	    case 0xB:
	    case 0x9:
	    case 0x8:
		int args[] = null;
		if (nodetype >= 0xA) {
		    int argcount = g_memory.readWord(nodeaddr + 5);
		    for (int i = 0; i < argcount; i++)
			args[i] = g_memory.readWord(nodeaddr + 9 + 4*i);
		}
		
		int indirectaddr = g_memory.readWord(nodeaddr + 1);
		if ((nodetype&1) == 1)
		    /* double indirect reference */
		    indirectaddr = g_memory.readWord(indirectaddr);
		/* indirect reference */
		if (!pushedtype11) {
		    make_callframe(0x11);
		    pushedtype11 = true;
		}
		int objtype = g_memory.readByte(indirectaddr);
		if (objtype == 0xE1) {
		    make_callframe(0x10, straddr, bitnum);
		    straddr = indirectaddr + 1;
		    bitnum = 0;
		}
		else if (objtype == 0xE0) {
		    if (current_iosys == iosys_Filter) {
			make_callframe(0x10, straddr, bitnum);
			stopdecoding = true;
		    }
		    do_streamE0string(nodeaddr+1, pushedtype11);
		}
		else { /* hope it's a function */
		    make_callframe(0x10, straddr, bitnum);
		    pc = indirectaddr;
		    make_callstack(args);
		    stopdecoding = true;
		}
		nodeaddr = rootaddr;
		break;
	    case 0:
		int mask = (1<<bitnum);
		int thebyte = g_memory.readByte(straddr);
		bitnum++;
		if (bitnum > 7) {
		    straddr++;
		    bitnum = 0;
		}
		if ((thebyte & mask) == 0) 
		    nodeaddr = g_memory.readWord(nodeaddr + 1);
		else
		    nodeaddr = g_memory.readWord(nodeaddr + 5);
		break;
		    
	    default:
		throw new GlulxException("Bogus nodetype encountered in string " + Integer.toHexString(nodetype));
	    }
	}
    }

    public void do_streamstring(int straddr) 
    throws GlulxException
    {
	int stringtype = g_memory.readByte(straddr++);
	switch (stringtype) {
	case 0xE0:
	    do_streamE0string(straddr);
	    break;
	case 0xE1:
	    do_streamE1string(straddr);
	    break;
	default:
	    throw new GlulxException("Illegal String Type " + Integer.toHexString(stringtype));
	}
    }

    public static void main(String argv[]) {
	Glulx myglulx = null;
	try {
	    FileInputStream in = new FileInputStream(argv[0]);
	    myglulx = new Glulx(in);
	    myglulx.testrun();
	}
	catch (Exception e) {
	    e.printStackTrace();
	    if (myglulx != null) {
		System.err.println("pc = " + Integer.toHexString(myglulx.pc));
		System.err.println("fp = " + Integer.toHexString(myglulx.fp));
		if (myglulx.g_stack != null)
		    System.err.println("sp = " + Integer.toHexString(myglulx.g_stack.getSP()));
	    }
	}
    }
    
    public static final int gestalt_GlulxVersion = 0;
    public static final int gestalt_TerpVersion = 1;
    public static final int gestalt_ResizeMem = 2;
    public static final int gestalt_Undo = 3;
    public static final int gestalt_IOSystem = 4;
    public static final int gestalt_IOSystem_Null = 0;
    public static final int gestalt_IOSystem_Filter = 1;
    public static final int gestalt_IOSystem_Glk = 2;
    public static final int gestalt_GlulxVersionValue = 0x00020000;
    public static final int gestalt_TerpVersionValue = 0x00000001;
    public static final int gestalt_ResizeMemValue = 0x00000001;
    public static final int gestalt_UndoValue = 0x00000000;
    public int gestalt(int selector, int arg)
    {
	switch (selector) {
	case gestalt_GlulxVersion: return gestalt_GlulxVersionValue;
	case gestalt_TerpVersion: return gestalt_TerpVersionValue;
	case gestalt_ResizeMem: return gestalt_ResizeMemValue;
	case gestalt_Undo: return gestalt_UndoValue;
	case gestalt_IOSystem: 
	    switch(arg) {
	    case gestalt_IOSystem_Null: 
	    case gestalt_IOSystem_Filter:
	    case gestalt_IOSystem_Glk:
		return 1;
	    default:
		return 0;
	    }
	}
	return 0;
    }
    
    void quit()
    {
	glk_exit();
    }

    Object toGlkOpaqueRef(int glulxref, int objclass) {
	if (glulxref == 0)
	    return null;
	return glkreg.glulxRefToGlkObject(glulxref, objclass);
    }

    int toGlulxOpaqueRef(Object glkobj, int objclass) {
	if (glkobj == null)
	  return 0;
	return glkreg.glkObjectToGlulxRef(glkobj, objclass);
    }
}

