package russotto.jglulx;
import java.io.PrintWriter;
import java.io.BufferedOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;

class Log {
    public final static int DEBUG_MIN = 1;
    public final static int DEBUG_EXTREME = 250;
    public final static int DEBUG = 500;
    public final static int DEBUG_RARE = 750;
    public final static int DEBUG_MAX = 1000;
    public final static int WARNING = 1001;
    public final static int ERROR = 2001;
    public final static int FATAL = 3001;

    public static boolean on = true;
    
    protected static int file_level = 0;
    protected static int console_level = 0;
    protected static PrintWriter log_writer;

    
    static {
	try {
	    log_writer = new PrintWriter(new BufferedOutputStream(new FileOutputStream("logfile")));    
	}
	catch (FileNotFoundException fnf) {
	    fnf.printStackTrace();
	    log_writer = null;
	}
	System.err.println("Log initialized");
    }

    public static void setConsoleLevel(int level)
    {
	console_level = level;
	System.err.println("console_level = " + console_level);
    }

    public static void setFileLevel(int level)
    {
	file_level = level;
    }
    
    protected static String levelToString(int level)
    {
	if (level < DEBUG_MIN)
	    return "Slime";
	else if (level  < WARNING)
	    return "Debug";
	else if (level < ERROR)
	    return "Warning";
	else if (level < FATAL)
	    return "Error";
	else
	    return "Fatal";
    }
    
    public static void log(int level, String routine, String message)
    {
	if (!on) return;

	String strlevel = levelToString(level);
	if (level >= console_level) {
	    System.err.print(strlevel);
	    System.err.print (": ");
	    System.err.print (routine);
	    System.err.print (" : ");
	    System.err.println(message);
	}
	if ((log_writer != null) && (level >= file_level)) {
	    log_writer.print(strlevel);
	    log_writer.print (": ");
	    log_writer.print (routine);
	    log_writer.print (" : ");
	    log_writer.println(message);
	}
    }

    public static void flush()
    {
	if (log_writer != null)
	    log_writer.flush();
    }

    public static boolean willLog(int level, String routine)
    {
	return on && ((level >= file_level) || (level >= console_level));
    }
}
